"""
This code is based on multiple sources:

https://github.com/rbgirshick/fast-rcnn
https://github.com/ultralytics/ultralytics
https://github.com/see--/keras-centernet
https://github.com/stefanopini/simple-HigherHRNet
"""

from enum import Enum
from typing import List

import cv2
import numpy as np

from picamera2 import Picamera2


def nms(dets: np.ndarray, scores: np.ndarray, iou_thres: float = 0.55, max_out_dets: int = 50) -> List[int]:
    """
    Perform Non-Maximum Suppression (NMS) on detected bounding boxes.

    Args:
        dets (np.ndarray): Array of bounding box coordinates of shape (N, 4) representing [y1, x1, y2, x2].
        scores (np.ndarray): Array of confidence scores associated with each bounding box.
        iou_thres (float, optional): IoU threshold for NMS. Default is 0.5.
        max_out_dets (int, optional): Maximum number of output detections to keep. Default is 300.

    Returns:
        List[int]: List of indices representing the indices of the bounding boxes to keep after NMS.

    """
    y1, x1 = dets[:, 0], dets[:, 1]
    y2, x2 = dets[:, 2], dets[:, 3]
    areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    order = scores.argsort()[::-1]

    keep = []
    while order.size > 0:
        i = order[0]
        keep.append(i)
        xx1 = np.maximum(x1[i], x1[order[1:]])
        yy1 = np.maximum(y1[i], y1[order[1:]])
        xx2 = np.minimum(x2[i], x2[order[1:]])
        yy2 = np.minimum(y2[i], y2[order[1:]])

        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)
        inter = w * h
        ovr = inter / (areas[i] + areas[order[1:]] - inter)

        inds = np.where(ovr <= iou_thres)[0]
        order = order[inds + 1]

    return keep[:max_out_dets]


def combined_nms(batch_boxes, batch_scores, iou_thres: float = 0.65, conf: float = 0.55, max_out_dets: int = 50):
    nms_results = []
    for boxes, scores in zip(batch_boxes, batch_scores):
        xc = np.argmax(scores, 1)
        xs = np.amax(scores, 1)
        x = np.concatenate([boxes, np.expand_dims(xs, 1), np.expand_dims(xc, 1)], 1)

        xi = xs > conf
        x = x[xi]

        x = x[np.argsort(-x[:, 4])[:8400]]
        scores = x[:, 4]
        x[..., :4] = convert_to_ymin_xmin_ymax_xmax_format(x[..., :4], BoxFormat.XC_YC_W_H)
        offset = x[:, 5] * 640
        boxes = x[..., :4] + np.expand_dims(offset, 1)

        # Original post-processing part
        valid_indexs = nms(boxes, scores, iou_thres=iou_thres, max_out_dets=max_out_dets)
        x = x[valid_indexs]
        nms_classes = x[:, 5]
        nms_bbox = x[:, :4]
        nms_scores = x[:, 4]

        nms_results.append((nms_bbox, nms_scores, nms_classes))

    return nms_results


def combined_nms_seg(batch_boxes, batch_scores, batch_masks, iou_thres: float = 0.5, conf: float = 0.001,
                     max_out_dets: int = 300):
    nms_results = []
    for boxes, scores, masks in zip(batch_boxes, batch_scores, batch_masks):
        # Compute maximum scores and corresponding class indices
        class_indices = np.argmax(scores, axis=1)
        max_scores = np.amax(scores, axis=1)
        detections = np.concatenate([boxes, np.expand_dims(max_scores, axis=1), np.expand_dims(class_indices, axis=1)],
                                    axis=1)

        # Swap the position of the two dimensions (32, 8400) to (8400, 32)
        masks = np.transpose(masks, (1, 0))
        # Filter out detections below the confidence threshold
        valid_detections = max_scores > conf

        if np.all(valid_detections is False):
            nms_results.append((np.ndarray(0), np.ndarray(0), np.ndarray(0), np.ndarray(0)))
        else:

            detections = detections[valid_detections]
            masks = masks[valid_detections]

            # Sort detections by score in descending order
            sorted_indices = np.argsort(-detections[:, 4])
            detections = detections[sorted_indices]
            masks = masks[sorted_indices]

            detections[..., :4] = convert_to_ymin_xmin_ymax_xmax_format(detections[..., :4], BoxFormat.XC_YC_W_H)

            # Perform class-wise NMS
            unique_classes = np.unique(detections[:, 5])
            final_indices = []

            for cls in unique_classes:
                cls_indices = np.where(detections[:, 5] == cls)[0]
                cls_boxes = detections[cls_indices, :4]
                cls_scores = detections[cls_indices, 4]
                cls_valid_indices = nms(cls_boxes, cls_scores, iou_thres=iou_thres, max_out_dets=max_out_dets)
                final_indices.extend(cls_indices[cls_valid_indices])

            final_indices = np.array(final_indices)
            final_detections = detections[final_indices]
            final_masks = masks[final_indices]

            # Extract class indices, bounding boxes, and scores
            nms_classes = final_detections[:, 5]
            nms_bbox = final_detections[:, :4]
            nms_scores = final_detections[:, 4]

            # Append results including masks
            nms_results.append((nms_bbox, nms_scores, nms_classes, final_masks))
    return nms_results


class BoxFormat(Enum):
    YMIM_XMIN_YMAX_XMAX = 'ymin_xmin_ymax_xmax'
    XMIM_YMIN_XMAX_YMAX = 'xmin_ymin_xmax_ymax'
    XMIN_YMIN_W_H = 'xmin_ymin_width_height'
    XC_YC_W_H = 'xc_yc_width_height'


def convert_to_ymin_xmin_ymax_xmax_format(boxes, orig_format: BoxFormat):
    """
    Changes the box from one format to another (XMIN_YMIN_W_H --> YMIM_XMIN_YMAX_XMAX )

    Also support in same format mode (returns the same format)

    :param boxes:
    :param orig_format:
    :return: box in format YMIM_XMIN_YMAX_XMAX
    """
    if len(boxes) == 0:
        return boxes
    elif orig_format == BoxFormat.YMIM_XMIN_YMAX_XMAX:
        return boxes
    elif orig_format == BoxFormat.XMIN_YMIN_W_H:
        boxes[:, 2] += boxes[:, 0]  # convert width to xmax
        boxes[:, 3] += boxes[:, 1]  # convert height to ymax
        boxes[:, 0], boxes[:, 1] = boxes[:, 1], boxes[:, 0].copy()  # swap xmin, ymin columns
        boxes[:, 2], boxes[:, 3] = boxes[:, 3], boxes[:, 2].copy()  # swap xmax, ymax columns
        return boxes
    elif orig_format == BoxFormat.XMIM_YMIN_XMAX_YMAX:
        boxes[:, 0], boxes[:, 1] = boxes[:, 1], boxes[:, 0].copy()  # swap xmin, ymin columns
        boxes[:, 2], boxes[:, 3] = boxes[:, 3], boxes[:, 2].copy()  # swap xmax, ymax columns
        return boxes
    elif orig_format == BoxFormat.XC_YC_W_H:
        new_boxes = np.copy(boxes)
        new_boxes[:, 0] = boxes[:, 1] - boxes[:, 3] / 2  # top left y
        new_boxes[:, 1] = boxes[:, 0] - boxes[:, 2] / 2  # top left x
        new_boxes[:, 2] = boxes[:, 1] + boxes[:, 3] / 2  # bottom right y
        new_boxes[:, 3] = boxes[:, 0] + boxes[:, 2] / 2  # bottom right x
        return new_boxes
    else:
        raise Exception("Unsupported boxes format")


def clip_boxes(boxes: np.ndarray, h: int, w: int) -> np.ndarray:
    """
    Clip bounding boxes to stay within the image boundaries.

    Args:
        boxes (numpy.ndarray): Array of bounding boxes in format [y_min, x_min, y_max, x_max].
        h (int): Height of the image.
        w (int): Width of the image.

    Returns:
        numpy.ndarray: Clipped bounding boxes.
    """
    boxes[..., 0] = np.clip(boxes[..., 0], a_min=0, a_max=h)
    boxes[..., 1] = np.clip(boxes[..., 1], a_min=0, a_max=w)
    boxes[..., 2] = np.clip(boxes[..., 2], a_min=0, a_max=h)
    boxes[..., 3] = np.clip(boxes[..., 3], a_min=0, a_max=w)
    return boxes


def scale_boxes(boxes: np.ndarray, h_image: int, w_image: int, h_model: int, w_model: int, preserve_aspect_ratio: bool,
                normalized: bool = True) -> np.ndarray:
    """
    Scale and offset bounding boxes based on model output size and original image size.

    Args:
        boxes (numpy.ndarray): Array of bounding boxes in format [y_min, x_min, y_max, x_max].
        h_image (int): Original image height.
        w_image (int): Original image width.
        h_model (int): Model output height.
        w_model (int): Model output width.
        preserve_aspect_ratio (bool): Whether to preserve image aspect ratio during scaling

    Returns:
        numpy.ndarray: Scaled and offset bounding boxes.
    """
    deltaH, deltaW = 0, 0
    H, W = h_model, w_model
    scale_H, scale_W = h_image / H, w_image / W

    if preserve_aspect_ratio:
        scale_H = scale_W = max(h_image / H, w_image / W)
        H_tag = int(np.round(h_image / scale_H))
        W_tag = int(np.round(w_image / scale_W))
        deltaH, deltaW = int((H - H_tag) / 2), int((W - W_tag) / 2)

    nh, nw = (H, W) if normalized else (1, 1)

    # Scale and offset boxes
    # [y_min, x_min, y_max, x_max].
    boxes[..., 0] = (boxes[..., 0] * nw - deltaW) * scale_W
    boxes[..., 1] = (boxes[..., 1] * nh - deltaH) * scale_H
    boxes[..., 2] = (boxes[..., 2] * nw - deltaW) * scale_W
    boxes[..., 3] = (boxes[..., 3] * nh - deltaH) * scale_H

    # Clip boxes
    boxes = clip_boxes(boxes, h_image, w_image)

    return boxes


def scale_coords(kpts: np.ndarray, h_image: int, w_image: int, h_model: int, w_model: int,
                 preserve_aspect_ratio: bool) -> np.ndarray:
    """
    Scale and offset keypoints based on model output size and original image size.

    Args:
        kpts (numpy.ndarray): Array of bounding keypoints in format [..., 17, 3]  where the last dim is (x, y, visible).
        h_image (int): Original image height.
        w_image (int): Original image width.
        h_model (int): Model output height.
        w_model (int): Model output width.
        preserve_aspect_ratio (bool): Whether to preserve image aspect ratio during scaling

    Returns:
        numpy.ndarray: Scaled and offset bounding boxes.
    """
    deltaH, deltaW = 0, 0
    H, W = h_model, w_model
    scale_H, scale_W = h_image / H, w_image / W

    if preserve_aspect_ratio:
        scale_H = scale_W = max(h_image / H, w_image / W)
        H_tag = int(np.round(h_image / scale_H))
        W_tag = int(np.round(w_image / scale_W))
        deltaH, deltaW = int((H - H_tag) / 2), int((W - W_tag) / 2)

    # Scale and offset boxes
    kpts[..., 0] = (kpts[..., 0] - deltaH) * scale_H
    kpts[..., 1] = (kpts[..., 1] - deltaW) * scale_W

    # Clip boxes
    kpts = clip_coords(kpts, h_image, w_image)

    return kpts


def clip_coords(kpts: np.ndarray, h: int, w: int) -> np.ndarray:
    """
    Clip keypoints to stay within the image boundaries.

    Args:
        kpts (numpy.ndarray): Array of bounding keypoints in format [..., 17, 3]  where the last dim is (x, y, visible).
        h (int): Height of the image.
        w (int): Width of the image.

    Returns:
        numpy.ndarray: Clipped bounding boxes.
    """
    kpts[..., 0] = np.clip(kpts[..., 0], a_min=0, a_max=h)
    kpts[..., 1] = np.clip(kpts[..., 1], a_min=0, a_max=w)
    return kpts


PARTS = {
    0: 'Nose',
    1: 'EyeL',
    2: 'EyeR',
    3: 'EarL',
    4: 'EarR',
    5: 'SholderL',
    6: 'SholderR',
    7: 'ElbowL',
    8: 'ElbowR',
    9: 'WristL',
    10: 'WristR',
    11: 'HipL',
    12: 'HipR',
    13: 'KneeL',
    14: 'KneeR',
    15: 'AnkleL',
    16: 'AnkleR'
}


class COCODrawer:
    def __init__(self, categories, imx500, needs_rescale_coords=True):
        self.categories = categories
        self.imx500 = imx500
        self.needs_rescale_coords = needs_rescale_coords

    def get_coords(self, annotation, metadata: dict, picam2: Picamera2, stream):
        if self.needs_rescale_coords:
            obj_scaled = self.imx500.convert_inference_coords(annotation, metadata, picam2, stream)
            x0 = obj_scaled.x
            y0 = obj_scaled.y
            x1 = x0 + obj_scaled.width
            y1 = y0 + obj_scaled.height
        else:
            y0, x0, y1, x1 = annotation
            y0 = max(0, y0)
            x0 = max(0, x0)
        return int(y0), int(x0), int(y1), int(x1)

    def draw_bounding_box(self, img, annotation, class_id, score, metadata: dict, picam2: Picamera2, stream):
        y0, x0, y1, x1 = self.get_coords(annotation, metadata, picam2, stream)
        text = f"{self.categories[int(class_id)]}:{score:.3f}"
        cv2.rectangle(img, (x0, y0), (x1, y1), (0, 0, 255), 2)
        cv2.putText(img, text, (x0, y0), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 255, 0), 1)

    def draw_keypoints(self, img, keypoints, min_confidence, metadata: dict, picam2: Picamera2, stream):
        def get_point(index):
            y0, x0 = keypoints[index][1], keypoints[index][0]
            y0, x0, _, _ = self.get_coords((y0, x0, y0 + 1, x0 + 1), metadata, picam2, stream)
            return x0, y0

        skeleton = [
            [0, 1], [0, 2], [1, 3], [2, 4],  # Head
            [5, 6], [5, 7], [7, 9], [6, 8],  # Arms
            [8, 10], [5, 11], [6, 12], [11, 12],  # Body
            [11, 13], [12, 14], [13, 15], [14, 16]  # Legs
        ]

        # Draw skeleton lines
        for connection in skeleton:
            start_point = get_point(connection[0])
            end_point = get_point(connection[1])
            start_confidence = keypoints[connection[0]][2]
            end_confidence = keypoints[connection[1]][2]
            if start_confidence < min_confidence or end_confidence < min_confidence:
                continue
            cv2.line(img, start_point, end_point, (255, 0, 0), 2)

        # Draw keypoints as colored circles
        for i in range(len(keypoints)):
            x, y = get_point(i)
            confidence = keypoints[i][2]
            if confidence < min_confidence:
                continue
            cv2.circle(img, (x, y), 3, (0, 255, 0), -1)
            label = f"{PARTS.get(i)}.{confidence:.3f}"
            cv2.putText(img, label, (x + 5, y + 15), cv2.FONT_HERSHEY_SIMPLEX, 0.25, (0, 255, 0), 1)

    def annotate_image(self, img, b, s, c, k, box_min_conf, kps_min_conf, metadata: dict, picam2: Picamera2, stream):
        for index, row in enumerate(b):
            if s[index] >= box_min_conf:
                self.draw_bounding_box(img, row, c[index], s[index], metadata, picam2, stream)
                if k is not None:
                    self.draw_keypoints(img, k[index], kps_min_conf, metadata, picam2, stream)

    def overlay_masks(self, picam2, masks, scores, colors, score_threshold=0.55, mask_threshold=0.5):
        overlay = np.zeros((masks.shape[1], masks.shape[2], 4), dtype=np.uint8)
        for idx, (mask, score) in enumerate(zip(masks, scores)):
            if score > score_threshold:  # Check if the score is above the threshold
                binary_mask = (mask > mask_threshold).astype(np.uint8)
                color = np.array(colors[idx][:3]) * 255  # Convert color to 0-255 scale
                overlay[binary_mask == 1, :3] = color
                overlay[binary_mask == 1, 3] = 127  # opacity
        picam2.set_overlay(overlay)


def softmax(x):
    y = np.exp(x - np.expand_dims(np.max(x, axis=-1), axis=-1))
    z = y / np.expand_dims(np.sum(y, axis=-1), axis=-1)
    return z


def crop_mask(masks, boxes):
    """
    It takes a mask and a bounding box, and returns a mask that is cropped to the bounding box

    Args:
      masks (numpy.ndarray): [h, w, n] tensor of masks
      boxes (numpy.ndarray): [n, 4] tensor of bbox coordinates in relative point form

    Returns:
      (numpy.ndarray): The masks are being cropped to the bounding box.
    """
    n, w, h = masks.shape
    x1, y1, x2, y2 = np.split(boxes[:, :, None], 4, 1)
    c = np.arange(h, dtype=np.float32)[None, None, :]  # rows shape(1,w,1)
    r = np.arange(w, dtype=np.float32)[None, :, None]  # cols shape(h,1,1)

    return masks * ((r >= x1) * (r < x2) * (c >= y1) * (c < y2))



"""Circular buffer"""

import collections
from threading import Lock

from .output import Output


class CircularOutput2(Output):
    """
    Circular buffer implementation, much like CircularOutput, but for general outputs.

    This means it can be used in conjunction with, for example, a PyavOutput to create time-shifted
    recordings of both video and audio straight to an mp4 file.

    Once the CircularOutput2 has been started, use the open_output method to start start recording
    a new output, and use close_output when finished. If the output has not been closed when the
    circular buffer is stopped, then the remainder of the buffer will be flush into the output.
    """

    def __init__(self, pts=None, buffer_duration_ms=5000):
        """Create a CircularOutput2."""
        super().__init__(pts=pts)
        # A note on locking. The lock is principally to protect outputframe, which is called by
        # the background encoder thread. Applications are going to call things like open_output,
        # close_output, start and stop. These only grab that lock for a short period of time to
        # manipulate _output, which controls whether outputframe will do anything.
        # THe application API does not have it's own lock, because there doesn't seem to be a
        # need to drive it from different threads (though we could add one if necessary).
        self._lock = Lock()
        if buffer_duration_ms < 0:
            raise RuntimeError("buffer_duration_ms may not be negative")
        self._buffer_duration_ms = buffer_duration_ms
        self._circular = collections.deque()
        self._output = None
        self._streams = []
        self._first_frame = True
        self._time_offset = 0

    @property
    def buffer_duration_ms(self):
        """Returns duration of the buffer in ms"""
        return self._buffer_duration_ms

    @buffer_duration_ms.setter
    def buffer_duration_ms(self, value):
        """Set buffer duration in ms, can even be changed dynamically"""
        with self._lock:
            self._buffer_duration_ms = value

    def open_output(self, output):
        """Open a new output object and start writing to it."""
        if self._output:
            raise RuntimeError("Underlying output must be closed first")

        output.start()
        # Some outputs (PyavOutput) may need to know about the encoder's streams.
        for encoder_stream, codec, kwargs in self._streams:
            output._add_stream(encoder_stream, codec, **kwargs)

        # Now it's OK for the background thread to output frames.
        with self._lock:
            self._output = output
            self._first_frame = True

    def close_output(self):
        """Close an output object."""
        if not self._output:
            raise RuntimeError("No underlying output has been opened")

        # After this, we guarantee that the background thread will never use the output.
        output = self._output
        with self._lock:
            self._output = None

        output.stop()

    def _flush(self, timestamp_now, output):
        # Flush out anything that is time-expired compared to timestamp_now.
        # If timestamp_now is None, flush everything.
        while self._circular and (front := self._circular[0]):
            frame, keyframe, timestamp, packet, audio = front

            if timestamp_now and timestamp_now - timestamp < self.buffer_duration_ms * 1000:
                break

            # We need to drop this entry, writing it out if we can.
            self._circular.popleft()

            if keyframe and not audio:
                if self._first_frame:
                    self._time_offset = timestamp
                self._first_frame = False

            if not self._first_frame and output:
                new_timestamp = timestamp - self._time_offset
                if new_timestamp >= 0:
                    output.outputframe(frame, keyframe, new_timestamp, packet, audio)

    def outputframe(self, frame, keyframe=True, timestamp=None, packet=None, audio=False):
        """Write frame to circular buffer"""
        with self._lock:
            if self._buffer_duration_ms == 0 or not self.recording:
                return

            # Add this new frame to the buffer and flush anything that is now expired.
            self._circular.append((frame, keyframe, timestamp, packet, audio))
            self._flush(timestamp, self._output)

    def start(self):
        """Start recording in the circular buffer."""
        with self._lock:
            if self.recording:
                raise RuntimeError("Circular output is running")
            self.recording = True

    def stop(self):
        """Close file handle and stop recording"""
        output = self._output
        with self._lock:
            if not self.recording:
                raise RuntimeError("Circular output was not started")
            self.recording = False
            self._output = None

        # At this point the background thread can't be using the circular buffer or the output,
        # so we can flush everything out.
        if output:
            self._flush(None, output)
            output.stop()

    def _add_stream(self, encoder_stream, codec_name, **kwargs):
        # Notice the PyavOutput of a stream that will be sending it packets to write out. It will need
        # to forward these whenever a new underlying output is opened.
        self._streams.append((encoder_stream, codec_name, kwargs))



import mmap
import multiprocessing as mp
import os
import queue
from collections import deque
from concurrent.futures import Future
from ctypes import CDLL, c_int, c_long, get_errno
from threading import Thread
from typing import Any, Callable

import numpy as np
from libcamera import ColorSpace, Transform
from PIL import Image

import picamera2


class Process:
    """
    Creates a new process which receives requests from the parent process.

    A wrapper for multiprocessing.Process that sends requests to a remote process
    and waits for the return values.

    This will timeout if no results are received for the timeout period, even if no requests are sent.

    Args:
        run: The function to run in the child process for each request
        picam2: The Picamera2 object
        init: The function to run in the child process to initialize the process
        timeout: The timeout for the return queue. It has a 1 second resolution.
    """

    def __init__(
            self,
            run: Callable[["RemoteRequest"], Any],
            picam2: picamera2.Picamera2,
            init: Callable[[], None] | None = None,
            timeout: float | None = 30):
        """
        Initializes the Process.

        Some configuration is copied from the Picamera2 object
        """
        self._send_queue = mp.Queue()
        self._return_queue = mp.Queue()

        self._requests_sent = deque()
        self._timeout = timeout
        self._thread = Thread(target=self._return_thread, args=(), daemon=True)
        self._thread.start()

        self._process = _RemoteProcess(self._send_queue, self._return_queue, picam2, run, init)

    def _return_thread(self):
        """Thread that passes the return values to the Future objects."""
        timeout_counter = 0
        while True:
            try:
                result = self._return_queue.get(timeout=1 if self._timeout is not None else None)
            except queue.Empty:
                if not bool(self._requests_sent):
                    timeout_counter = 0
                    continue
                timeout_counter += 1
                if timeout_counter >= self._timeout:
                    _, future = self._requests_sent.popleft()
                    future.set_exception(TimeoutError("No result received for timeout period"))
                continue

            timeout_counter = 0

            if not bool(self._requests_sent):
                break
            request, future = self._requests_sent.popleft()
            future.set_result(result)
            request.release()

    def send(self, request: picamera2.request.CompletedRequest, **kwargs):
        """
        Sends a request to the child process.

        The request is converted to a RemoteRequest object and sent to the child process.
        Returns a Future object that can be used to wait for the result.
        """
        future = Future()
        request.acquire()
        remote = RemoteRequest(request, **kwargs)
        self._send_queue.put(remote)
        self._requests_sent.append((request, future))
        return future

    def close(self):
        """Closes the Process."""
        self._send_queue.put("CLOSE")
        self._thread.join()
        self._process.join()


class _RemoteProcess(mp.Process):
    """
    Starts a new process which receives requests from the parent process.

    The process receives requests from the parent process and runs the given function.
    Then it returns the results back to the parent process.
    """

    def __init__(
            self,
            send_queue: mp.Queue,
            return_queue: mp.Queue,
            picam2: picamera2.Picamera2,
            run: Callable[["RemoteRequest"], Any],
            init: Callable[[], None],
            *args,
            **kwargs):
        super().__init__(*args, **kwargs)
        self._send_queue = send_queue
        self._return_queue = return_queue
        self._return_result = False  # Whether the parent is expecting a result

        self._pid_fd = None
        self._picam2_pid = os.getpid()
        self._array_cache = {}
        self._buffer_cache = {}
        self._request = None
        self._helper = picamera2.request.Helpers(_RemotePicamera2(picam2))

        self._run = run
        self._init_func = init
        self.start()

    def _child_init(self):
        """Initialization that runs in the child process."""
        self._syscall = CDLL(None, use_errno=True).syscall
        self._syscall.argtypes = [c_long]
        if self._init_func is not None:
            self._init_func()

    def run(self):
        """Main loop of the child process."""
        self._child_init()
        while (request := self._capture_request()) is not None:
            value = self._run(request, **request._kwargs)
            self._return_request(value)

    def _map_fd(self, picam2_fd: int):
        """Maps a file descriptor from the parent process to the child process."""
        if self._pid_fd is None:
            self._pid_fd = os.pidfd_open(self._picam2_pid)
        fd = self._syscall(438, c_int(self._pid_fd), c_int(picam2_fd), c_int(0))  # 438 is pidfd_getfd
        if fd == -1:
            errno = get_errno()
            raise OSError(errno, os.strerror(errno))
        return fd

    def _capture_request(self):
        """Captures a request from the parent process."""
        if self._return_result:
            self._return_queue.put(None)
        self._return_result = True

        msg = self._send_queue.get()
        if msg == "CLOSE":
            self._return_queue.put(None)
            return None

        msg._deserialize(self)
        self._request = msg
        return msg

    def _return_request(self, return_value: Any):
        """Returns a request to the parent process."""
        if self._request is None:
            raise ValueError("No request to return")

        if self._request._array_ref_count > 0:
            raise ValueError("Cannot return request while arrays are still in use")

        self._request = None
        self._return_queue.put(return_value)
        self._return_result = False


class RemoteMappedArray:
    """
    A mapped array that is created in the child process.

    Args:
        request: The RemoteRequest object
        stream_name: The name of the stream to map
    """

    def __init__(self, request: "RemoteRequest", stream_name: str):
        """Initializes the RemoteMappedArray."""
        self._request = request
        self._stream_name = stream_name
        self._array = None

    def __enter__(self):
        self._request._array_ref_count += 1
        self._array = self._request._arrays[self._stream_name]
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._request._array_ref_count -= 1
        if self._request._array_ref_count < 0:
            raise ValueError("Array reference count is less than 0")
        self._array = None

    @property
    def array(self):
        """Access the array."""
        return self._array


class RemoteRequest:
    """
    A request that is sent to the child process.

    Args:
        request: The CompletedRequest object
    """

    def __init__(self, request: picamera2.request.CompletedRequest, **kwargs):
        """Initializes and serializes the request."""
        self.request = request
        self._metadata = request.get_metadata()
        self.config = request.config.copy()
        t = self.config["transform"]
        self.config["transform"] = (t.hflip, t.vflip, t.transpose)
        c = self.config["colour_space"]
        self.config["colour_space"] = (c.primaries, c.transferFunction, c.ycbcrEncoding, c.range)
        self._array_ref_count = 0

        self._buffers_fd = {}
        for name in self.request.picam2.stream_map:
            if self.request.picam2.stream_map[name] is None:
                self._buffers_fd[name] = None
            else:
                self._buffers_fd[name] = self._serialize_stream(name)

        self._kwargs = kwargs

        del self.request

    def _serialize_stream(self, stream_name: str):
        """Serializes a stream to a file descriptor and length."""
        plane = self.request.request.buffers[self.request.picam2.stream_map[stream_name]].planes[0]
        fd = plane.fd
        length = plane.length
        return fd, length

    def _deserialize(self, process: _RemoteProcess):
        """Deserializes the request."""
        self._arrays = {}
        self._buffers = {}
        self._process = process
        for name in self._buffers_fd:
            if self._buffers_fd[name] is not None:
                pid_fd, length = self._buffers_fd[name]
                if pid_fd in process._array_cache:
                    self._arrays[name] = process._array_cache[pid_fd]
                    self._buffers[name] = process._buffer_cache[pid_fd]
                else:
                    fd = process._map_fd(pid_fd)
                    buffer = self._deserialize_buffer(fd, length)
                    self._arrays[name] = self._create_array(buffer, name)
                    self._buffers[name] = buffer
                    process._array_cache[pid_fd] = self._arrays[name]
                    process._buffer_cache[pid_fd] = self._buffers[name]

        self.config["transform"] = Transform(
            self.config["transform"][0],
            self.config["transform"][1],
            self.config["transform"][2])
        self.config["colour_space"] = ColorSpace(
            self.config["colour_space"][0],
            self.config["colour_space"][1],
            self.config["colour_space"][2],
            self.config["colour_space"][3])

    def _deserialize_buffer(self, fd: int, length: int):
        """Deserializes a buffer."""
        mem = mmap.mmap(fd, length, mmap.MAP_SHARED, mmap.PROT_READ | mmap.PROT_WRITE)
        return mem

    def _create_array(self, mem, name: str):
        """Creates an array from a buffer."""
        array = np.array(mem, copy=False, dtype=np.uint8)
        config = self.config[name]
        width, height = config["size"]
        stride = config["stride"]
        format = config["format"]
        if format == "YUV420":
            return array.reshape((height + height // 2, stride))
        array = array.reshape((height, stride))
        if format in ("RGB888", "BGR888"):
            return array[:, :width * 3].reshape((height, width, 3))
        elif format in ("XBGR8888", "XRGB8888"):
            return array[:, :width * 4].reshape((height, width, 4))
        return array

    def get_metadata(self):
        """Access the metadata."""
        return self._metadata

    def get_config(self):
        """Access the config."""
        return self.config

    def make_buffer(self, stream_name: str):
        """Access the buffer of a stream."""
        return np.array(self._buffers[stream_name], dtype=np.uint8)

    def make_array(self, stream_name: str):
        """Create a numpy array a stream."""
        with RemoteMappedArray(self, stream_name) as m:
            if m.array.data.c_contiguous:
                return np.copy(m.array)
            else:
                return np.ascontiguousarray(m.array)

    def make_image(self, stream_name: str):
        """Create an Image from a stream."""
        config = self.config.get(stream_name, None)
        if config is None:
            raise RuntimeError(f"Stream {stream_name!r} is not defined")

        fmt = config["format"]
        mode = self._process._helper._get_pil_mode(fmt)

        with RemoteMappedArray(self, stream_name) as m:
            shape = m.array.shape
            stride = m.array.strides[0]
            if mode == "RGBX":
                array = np.copy(m.array)
                stride = array.strides[0]
            elif not m.array.data.c_contiguous:
                array = m.array.base
            img = Image.frombuffer("RGB", (shape[1], shape[0]), array, "raw", mode, stride, 1)
        return img

    def save(self, file_output: str, name: str = "main", format_str: str | None = None, exif_data: dict | None = None):
        """Save an image to a file."""
        img = self.make_image(name)
        self._process._helper.save(img, self.get_metadata(), file_output, format_str, exif_data)

    def save_dng(self, file_output: str, name: str = "raw"):
        """Save a the image to a DNG file."""
        buffer = self.make_buffer(name)
        self._process._helper.save_dng(buffer, self.get_metadata(), self.config[name], file_output)


class _FakeObject:
    """A fake object that is used to access the camera properties in the child process."""

    def __init__(self, **kwargs):
        """Initializes the FakeObject."""
        self.__dict__.update(kwargs)


class _RemotePicamera2:
    """
    A fake object that looks like a Picamera2 object in the child process.

    Some of the helper functions get their configuration from the picam2 object.
    This stores the properties that are needed in the child process.
    """

    def __init__(self, picam2: picamera2.Picamera2):
        self.camera = _FakeObject(id=picam2.camera.id)
        self.options = {}
        if "compress_level" in picam2.options:
            self.options["compress_level"] = picam2.options["compress_level"]
        if "quality" in picam2.options:
            self.options["quality"] = picam2.options["quality"]

        self.camera_properties = {}
        if "Model" in picam2.camera_properties:
            self.camera_properties["Model"] = picam2.camera_properties["Model"]


class Pool:
    """
    A pool of processes that are used to process the images.

    This can be used as a context manager to automatically close the pool.

    Args:
        run: The function to run in each child process for each request
        count: The number of processes to create
        picam2: The Picamera2 object
        init: The function to run in each child process to initialize the process
    """

    def __init__(
            self,
            run: Callable[["RemoteRequest"], Any],
            count: int,
            picam2: picamera2.Picamera2,
            init: Callable[[], None] | None = None,
            timeout: float | None = 30):
        """Initializes the Pool."""
        self._processes = [Process(run, picam2, init, timeout) for _ in range(count)]
        self._process_count = count
        self._process_index = 0  # The preferred process for the next request

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    def send(self, request: picamera2.request.CompletedRequest, **kwargs):
        """Sends a request to the child process."""
        process = min(
            enumerate(self._processes),
            key=lambda p: (
                len(p[1]._requests_sent),
                (p[0] + self._process_index) % self._process_count
            )
        )[1]
        self._process_index = (self._process_index + 1) % self._process_count

        return process.send(request, **kwargs)

    def close(self):
        """Closes the Pool."""
        for process in self._processes:
            process.close()



try:
    # Hailo requires hailo_platform package, which may not be installed on non-Hailo platforms.
    from .hailo import Hailo
except ModuleNotFoundError:
    pass
from .imx500 import IMX500
from .imx708 import IMX708



"""
Efficientdet postprocessing

This code is based on:
https://github.com/google/automl/tree/master/efficientdet
"""

from typing import Tuple

import numpy as np

from picamera2.devices.imx500.postprocess import (
    BoxFormat, convert_to_ymin_xmin_ymax_xmax_format, nms)
from picamera2.devices.imx500.postprocess_yolov5 import coco80_to_coco91

default_box_variance = [1.0, 1.0, 1.0, 1.0]
default_aspect_ratios = [1.0, 2.0, 0.5]


def postprocess_efficientdet_lite0_detection(outputs: Tuple[np.ndarray, np.ndarray, np.ndarray],
                                             anchor_scale=3,
                                             min_level=3,
                                             max_level=7,
                                             box_variance=default_box_variance,
                                             model_input_shape=(320, 320),
                                             min_wh=2,
                                             max_wh=7680,
                                             conf_thres: float = 0.001,
                                             iou_thres: float = 0.65,
                                             max_nms_dets: int = 5000,
                                             max_out_dets: int = 1000):
    H, W = model_input_shape
    ############################################################
    # Box decoding
    ############################################################
    outputs_decoded = box_decoding_edetlite(output_annotations=outputs,
                                            H=H,
                                            W=W,
                                            anchor_scale=anchor_scale,
                                            min_level=min_level,
                                            max_level=max_level,
                                            box_variance=box_variance)

    classes = outputs[0]
    num_categories = classes.shape[-1]

    ############################################################
    # Post processing for each input image
    ############################################################
    # Note: outputs_decoded shape is [Batch,num_anchors*Detections,(4+1+num_categories)]
    post_processed_outputs = []
    for _, x in enumerate(outputs_decoded):
        # ----------------------------------------
        # Filter by score and width-height
        # ----------------------------------------
        scores = x[..., 4]
        wh = x[..., 2:4]
        valid_indexs = (scores > conf_thres) & ((wh > min_wh).any(1)) & ((wh < max_wh).any(1))
        x = x[valid_indexs]

        # ----------------------------------------
        # Taking Best class only
        # ----------------------------------------
        x[..., 5:] *= x[..., 4:5]  # compute confidence per class (class_score * object_score)
        conf = np.max(x[:, 5:], axis=1, keepdims=True)
        classes_id = np.argmax(x[:, 5:], axis=1, keepdims=True)

        # Change boxes format from [x_c,y_c,w,h] to [y_min,x_min,y_max,x_max]
        boxes = convert_to_ymin_xmin_ymax_xmax_format(x[..., :4], BoxFormat.XC_YC_W_H)
        x = np.concatenate((boxes, conf, classes_id), axis=1)[conf.reshape(-1) > conf_thres]

        # --------------------------- #
        # NMS
        # --------------------------- #
        x = x[np.argsort(-x[:, 4])[:max_nms_dets]]  # sort by confidence from high to low
        offset = x[..., 5:6] * np.maximum(H, W)
        boxes_offset, scores = x[..., :4] + offset, x[..., 4]  # boxes with offset by class
        valid_indexs = nms(dets=boxes_offset, scores=scores, iou_thres=iou_thres, max_out_dets=max_out_dets)
        x = x[valid_indexs]

        boxes = x[..., :4]

        # --------------------------- #
        # Classes process
        # --------------------------- #
        # convert classes from coco80 to coco91 to match labels
        classes = coco80_to_coco91(x[..., 5]) if num_categories == 80 else x[..., 5]
        classes -= 0

        # --------------------------- #
        # Scores
        # --------------------------- #
        scores = x[..., 4]

        # Add result
        post_processed_outputs.append({'boxes': boxes, 'classes': classes, 'scores': scores})

    return post_processed_outputs[0]['boxes'], post_processed_outputs[0]['scores'], post_processed_outputs[0]['classes']


def box_decoding_edetlite(output_annotations,
                          H=320,
                          W=320,
                          anchor_scale=3,
                          min_level=3,
                          max_level=7,
                          box_variance=default_box_variance):
    # -----------------------------------------------
    # EfficientDetLite detection post processing
    # -----------------------------------------------
    # Note: 'output_annotations' is expected to be a list of 2 feature maps with shapes:
    # [0] : [Batch,Detections,num_categories]
    # [1] : [Batch,Detections,4]
    classes = output_annotations[0]
    boxes = output_annotations[1]
    classes = 1 / (1 + np.exp(-classes))  # sigmoid
    scores = np.ones((*boxes.shape[:-1], 1))  # Add default object scores of 1.0

    # Combine tensors
    outputs = np.concatenate((boxes, scores, classes), axis=2)

    # Box decoding
    # Anchor boxes format: [y_min, x_min, y_max, x_max] normalized

    # Extract feature map sizes
    strides = [2 ** i for i in range(max_level + 1)]
    featmap_sizes = [(np.ceil(H / stride), np.ceil(W / stride)) for stride in strides]

    # Generate priors
    batch_size = outputs.shape[0]
    anchors = generate_anchors_EDETLITE(batch_size=batch_size,
                                        featmap_sizes=featmap_sizes,
                                        H=H,
                                        W=W,
                                        anchor_scale=anchor_scale,
                                        min_level=min_level,
                                        max_level=max_level)

    # Decode bboxes
    y_c_anchors = (anchors[..., 0:1] + anchors[..., 2:3]) / 2
    x_c_anchors = (anchors[..., 1:2] + anchors[..., 3:4]) / 2
    ha = anchors[..., 2:3] - anchors[..., 0:1]
    wa = anchors[..., 3:4] - anchors[..., 1:2]

    # Output Box format: [x_c, y_c, w, h]
    pred_boxes = outputs[..., :4]
    y_c = pred_boxes[..., 0:1] * box_variance[0] * ha + y_c_anchors
    x_c = pred_boxes[..., 1:2] * box_variance[1] * wa + x_c_anchors
    h = np.exp(pred_boxes[..., 2:3] * box_variance[2]) * ha
    w = np.exp(pred_boxes[..., 3:4] * box_variance[3]) * wa
    outputs[..., 0:1] = x_c
    outputs[..., 1:2] = y_c
    outputs[..., 2:3] = w
    outputs[..., 3:4] = h
    return outputs


def generate_anchors_EDETLITE(batch_size,
                              featmap_sizes,
                              H=320,
                              W=320,
                              anchor_scale=3,
                              min_level=3,
                              max_level=7,
                              aspect_ratios=default_aspect_ratios):
    """Generate configurations of anchor boxes."""
    anchor_scales = [anchor_scale] * (max_level - min_level + 1)
    num_scales = len(aspect_ratios)
    anchor_configs = {}
    for level in range(min_level, max_level + 1):
        anchor_configs[level] = []
        for scale_octave in range(num_scales):
            for aspect in aspect_ratios:
                anchor_configs[level].append(
                    ((featmap_sizes[0][0] / float(featmap_sizes[level][0]),
                      featmap_sizes[0][1] / float(featmap_sizes[level][1])),
                     scale_octave / float(num_scales), aspect,
                     anchor_scales[level - min_level]))

    """Generates multiscale anchor boxes."""
    boxes_all = []
    for _, configs in anchor_configs.items():
        boxes_level = []
        for config in configs:
            stride, octave_scale, aspect, anchor_scale = config
            base_anchor_size_x = anchor_scale * stride[1] * 2 ** octave_scale
            base_anchor_size_y = anchor_scale * stride[0] * 2 ** octave_scale
            if isinstance(aspect, list):
                aspect_x, aspect_y = aspect
            else:
                aspect_x = np.sqrt(aspect)
                aspect_y = 1.0 / aspect_x
            anchor_size_x_2 = base_anchor_size_x * aspect_x / 2.0
            anchor_size_y_2 = base_anchor_size_y * aspect_y / 2.0

            x = np.arange(stride[1] / 2, W, stride[1])
            y = np.arange(stride[0] / 2, H, stride[0])
            xv, yv = np.meshgrid(x, y)
            xv = xv.reshape(-1)
            yv = yv.reshape(-1)

            boxes = np.vstack((yv - anchor_size_y_2, xv - anchor_size_x_2,
                               yv + anchor_size_y_2, xv + anchor_size_x_2))
            boxes = np.swapaxes(boxes, 0, 1)
            boxes_level.append(np.expand_dims(boxes, axis=1))

        # concat anchors on the same level to the shape Batch x Detections x 4
        boxes_level = np.concatenate(boxes_level, axis=1).reshape([1, -1, 4])
        boxes_level = np.repeat(boxes_level, batch_size, axis=0)
        boxes_all.append(boxes_level)

    anchor_boxes = np.concatenate(boxes_all, axis=1)
    return anchor_boxes



"""
Yolov5 postprocessing

This code is based on:
https://github.com/ultralytics/ultralytics
"""
from typing import List

import cv2
import numpy as np

from picamera2.devices.imx500.postprocess import (
    BoxFormat, convert_to_ymin_xmin_ymax_xmax_format, nms)

default_anchors = [[10, 13, 16, 30, 33, 23],
                   [30, 61, 62, 45, 59, 119],
                   [116, 90, 156, 198, 373, 326]]
default_strides = [8, 16, 32]


def postprocess_yolov5_detection(outputs: List[np.ndarray],
                                 model_input_shape=(640, 640),
                                 num_categories=80,
                                 min_wh=2,
                                 max_wh=7680,
                                 conf_thres: float = 0.001,
                                 iou_thres: float = 0.65,
                                 max_nms_dets: int = 5000,
                                 max_out_dets: int = 1000):
    H, W = model_input_shape
    ############################################################
    # Box decoding
    ############################################################
    outputs_decoded = box_decoding_yolov5n(tensors=outputs, num_categories=num_categories, H=H, W=W)

    ############################################################
    # Post processing for each input image
    ############################################################
    # Note: outputs_decoded shape is [Batch,num_anchors*Detections,(4+1+num_categories)]
    post_processed_outputs = []
    for _, x in enumerate(outputs_decoded):
        # ----------------------------------------
        # Filter by score and width-height
        # ----------------------------------------
        scores = x[..., 4]
        wh = x[..., 2:4]
        valid_indexs = (scores > conf_thres) & ((wh > min_wh).any(1)) & ((wh < max_wh).any(1))
        x = x[valid_indexs]

        # ----------------------------------------
        # Taking Best class only
        # ----------------------------------------
        x[..., 5:] *= x[..., 4:5]  # compute confidence per class (class_score * object_score)
        conf = np.max(x[:, 5:], axis=1, keepdims=True)
        classes_id = np.argmax(x[:, 5:], axis=1, keepdims=True)

        # Change boxes format from [x_c,y_c,w,h] to [y_min,x_min,y_max,x_max]
        boxes = convert_to_ymin_xmin_ymax_xmax_format(x[..., :4], BoxFormat.XC_YC_W_H)
        x = np.concatenate((boxes, conf, classes_id), axis=1)[conf.reshape(-1) > conf_thres]

        # --------------------------- #
        # NMS
        # --------------------------- #
        x = x[np.argsort(-x[:, 4])[:max_nms_dets]]  # sort by confidence from high to low
        offset = x[..., 5:6] * np.maximum(H, W)
        boxes_offset, scores = x[..., :4] + offset, x[..., 4]  # boxes with offset by class
        valid_indexs = nms(dets=boxes_offset, scores=scores, iou_thres=iou_thres, max_out_dets=max_out_dets)
        x = x[valid_indexs]

        boxes = x[..., :4]
        # --------------------------- #
        # Classes process
        # --------------------------- #
        # convert classes from coco80 to coco91 to match labels
        classes = coco80_to_coco91(x[..., 5]) if num_categories == 80 else x[..., 5]
        classes -= 1

        # --------------------------- #
        # Scores
        # --------------------------- #
        scores = x[..., 4]

        # Add result
        post_processed_outputs.append({'boxes': boxes, 'classes': classes, 'scores': scores})

    return post_processed_outputs[0]['boxes'], post_processed_outputs[0]['scores'], post_processed_outputs[0]['classes']


def box_decoding_yolov5n(tensors,
                         num_categories=80,
                         H=640,
                         W=640,
                         anchors=default_anchors,
                         strides=default_strides):
    # Tensors box format: [x_c, y_c, w, h]
    no = num_categories + 5  # number of outputs per anchor
    nl = len(anchors)  # number of detection layers
    na = len(anchors[0]) // 2  # number of anchors
    anchor_grid = np.reshape(np.array(anchors), [nl, 1, -1, 1, 2])
    anchor_grid = anchor_grid.astype(np.float32)
    z = []
    for i in range(nl):
        ny, nx = H // strides[i], W // strides[i]
        xv, yv = np.meshgrid(np.arange(nx), np.arange(ny))
        grid = np.reshape(np.stack([xv, yv], 2), [1, 1, ny * nx, 2]).astype(np.float32)

        y = tensors[i]
        y = np.transpose(y, [0, 2, 1, 3])
        xy = (y[..., 0:2] * 2 - 0.5 + grid) * strides[i]  # xy
        wh = (y[..., 2:4] * 2) ** 2 * anchor_grid[i]

        # Output box format: [x_c, y_c, w, h]
        y = np.concatenate([xy, wh, y[..., 4:]], -1)
        z.append(np.reshape(y, [-1, na * ny * nx, no]))

    return np.concatenate(z, 1)


# same as in preprocess but differs in h/w location
def scale_boxes(boxes: np.ndarray, h_image: int, w_image: int, h_model: int, w_model: int,
                preserve_aspect_ratio: bool) -> np.ndarray:
    """
    Scale and offset bounding boxes based on model output size and original image size.

    Args:
        boxes (numpy.ndarray): Array of bounding boxes in format [y_min, x_min, y_max, x_max].
        h_image (int): Original image height.
        w_image (int): Original image width.
        h_model (int): Model output height.
        w_model (int): Model output width.
        preserve_aspect_ratio (bool): Whether to preserve image aspect ratio during scaling

    Returns:
        numpy.ndarray: Scaled and offset bounding boxes.
    """
    deltaH, deltaW = 0, 0
    H, W = h_model, w_model
    scale_H, scale_W = h_image / H, w_image / W

    if preserve_aspect_ratio:
        scale_H = scale_W = max(h_image / H, w_image / W)
        H_tag = int(np.round(h_image / scale_H))
        W_tag = int(np.round(w_image / scale_W))
        deltaH, deltaW = int((H - H_tag) / 2), int((W - W_tag) / 2)

    # Scale and offset boxes
    boxes[..., 0] = (boxes[..., 0] - deltaH) * scale_H
    boxes[..., 1] = (boxes[..., 1] - deltaW) * scale_W
    boxes[..., 2] = (boxes[..., 2] - deltaH) * scale_H
    boxes[..., 3] = (boxes[..., 3] - deltaW) * scale_W

    # Clip boxes
    boxes = clip_boxes(boxes, h_image, w_image)

    return boxes


# same as in preprocess but differs in h/w location
def clip_boxes(boxes: np.ndarray, h: int, w: int) -> np.ndarray:
    """
    Clip bounding boxes to stay within the image boundaries.

    Args:
        boxes (numpy.ndarray): Array of bounding boxes in format [y_min, x_min, y_max, x_max].
        h (int): Height of the image.
        w (int): Width of the image.

    Returns:
        numpy.ndarray: Clipped bounding boxes.
    """
    boxes[..., 0] = np.clip(boxes[..., 0], a_min=0, a_max=h)
    boxes[..., 1] = np.clip(boxes[..., 1], a_min=0, a_max=w)
    boxes[..., 2] = np.clip(boxes[..., 2], a_min=0, a_max=h)
    boxes[..., 3] = np.clip(boxes[..., 3], a_min=0, a_max=w)
    return boxes


def _normalize_coordinates(boxes, orig_width, orig_height, boxes_format):
    """
    Gets boxes in the original images values and normalize them to be between 0 to 1

    :param boxes:
    :param orig_width: original image width
    :param orig_height: original image height
    :param boxes_format: if the boxes are in XMIN_YMIN_W_H or YMIM_XMIN_YMAX_XMAX format
    :return:
    """
    if len(boxes) == 0:
        return boxes
    elif _are_boxes_normalized(boxes):
        return boxes
    boxes[:, 0] = np.divide(boxes[:, 0], orig_height)
    boxes[:, 1] = np.divide(boxes[:, 1], orig_width)
    boxes[:, 2] = np.divide(boxes[:, 2], orig_height)
    boxes[:, 3] = np.divide(boxes[:, 3], orig_width)
    return boxes


def _are_boxes_normalized(boxes):
    if len(boxes) == 0:
        return True  # it doesn't matter
    if max(boxes[0]) > 1:
        return False
    return True


def apply_normalization(boxes, orig_width, orig_height, boxes_format):
    if _are_boxes_normalized(boxes):
        return boxes
    return _normalize_coordinates(boxes, orig_width, orig_height, boxes_format)


# Locate at tutorials
def coco80_to_coco91(x):  # converts 80-index to 91-index
    coco91Indexs = np.array(
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 27, 28, 31, 32, 33, 34,
         35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62,
         63, 64, 65, 67, 70, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 84, 85, 86, 87, 88, 89, 90])

    return coco91Indexs[x.astype(np.int32)]


def yolov5n_preprocess(img):
    # AspectPreservingResizeWithPad
    new_height = 640
    new_width = 640
    pad_value = 114
    resize_method = 3  # area
    resize_ratio = max(img.shape[0] / new_height, img.shape[1] / new_width)
    height_tag = int(np.round(img.shape[0] / resize_ratio))
    width_tag = int(np.round(img.shape[1] / resize_ratio))
    pad_values = ((int((new_height - height_tag) / 2), int((new_height - height_tag) / 2 + 0.5)),
                  (int((new_width - width_tag) / 2), int((new_width - width_tag) / 2 + 0.5)),
                  (0, 0))

    resized_img = cv2.resize(img, (width_tag, height_tag), interpolation=resize_method)
    padded_img = np.pad(resized_img, pad_values, constant_values=pad_value)

    # Normalize
    mean = 0
    std = 255
    normalized_img = (padded_img - mean) / std

    return normalized_img



from fractions import Fraction

import av

from .output import Output


class PyavOutput(Output):
    """
    The PyavOutput class outputs an encoded video, and optionally audio, stream using PyAV.

    PyAv is a Python interface to libav, used by FFmpeg, and therefore can accept many different output
    types and destinations, in the same way as FFmpeg.

    The PyavOutput calls directly into libav through its Python layer, and does not pipe encoded frames
    out to a separate process like the FfmpegOutput. The PyavOutput integration means we can pass precise
    timestamps, and are not subject to the hazards of FFmpeg re-timestamping everything as it gets piped
    back in.
    """

    def __init__(self, output_name, format=None, pts=None, options=None):
        super().__init__(pts=pts)
        self._output_name = output_name
        self._format = format
        self._streams = {}
        self._container = None
        self._options = options
        # A user can set this to get notifications of failures.
        self.error_callback = None
        self.needs_add_stream = True

    def _add_stream(self, encoder_stream, codec_name, **kwargs):
        # The output container that does the muxing needs to know about the streams for which packets
        # will be sent to it. It literally needs to copy them for the output container.
        stream = self._container.add_stream(codec_name, **kwargs)

        if codec_name == "mjpeg":
            # Well, this is nasty. MJPEG seems to need this.
            stream.codec_context.color_range = 2  # JPEG (full range)

        self._streams[encoder_stream] = stream

    def start(self):
        """Start the PyavOutput."""
        self._container = av.open(self._output_name, "w", format=self._format, options=self._options)
        super().start()

    def stop(self):
        """Stop the PyavOutput."""
        super().stop()
        if self._container:
            try:
                self._container.close()
            except Exception:
                pass
            self._container = None
            self._streams = {}

    def outputframe(self, frame, keyframe=True, timestamp=None, packet=None, audio=False):
        """Output an encoded frame using PyAv."""
        if self.recording and self._container:
            # We must make a packet that looks like it came from our own container's version of the stream.
            if not packet:
                # No packet present. It must have come from a video encoder that isn't using libav, so make one up.
                packet = av.Packet(frame)
                packet.dts = timestamp
                packet.pts = timestamp
                packet.time_base = Fraction(1, 1000000)
                packet.is_keyframe = keyframe
                packet.stream = self._streams["video"]
            else:
                # We can copy the packet we are given, updating the stream to be our version, and amending
                # the timestamp in case that has been changed.
                new_packet = av.Packet(packet)
                if packet.stream not in self._streams:
                    raise RuntimeError("Stream not found in PyavOutput")
                new_packet.stream = self._streams[packet.stream]
                new_packet.dts = timestamp
                new_packet.pts = timestamp
                new_packet.time_base = Fraction(1, 1000000)
                new_packet.is_keyframe = keyframe
                packet = new_packet

            try:
                self._container.mux(packet)
            except Exception as e:
                try:
                    self._container.close()
                except Exception:
                    pass
                self._container = None
                if self.error_callback:
                    self.error_callback(e)

            self.outputtimestamp(timestamp)



import importlib
from enum import Enum
from types import ModuleType


class _QT_BINDING(Enum):
    PyQt5 = 'PyQt5'
    PyQt6 = 'PyQt6'
    PySide2 = 'PySide2'
    PySide6 = 'PySide6'


def _get_qt_modules(qt_module: _QT_BINDING) -> tuple[ModuleType, ModuleType, ModuleType]:
    """
    Gets the qt modules for the given qt binding

    Parameters
    ----------
    qt_module : _QT_BINDING
        The module we want to get Qt QtCore, QtGui, and QtWidgets modules from

    Returns
    -------
    out : tuple[ModuleType, ModuleType, ModuleType]
        A tuple containing the QtCore, QtGui, and QtWidgets modules respectively, from the given python qt module

    Notes
    -----
    - Assigns alternate attributes for Signals and Slots as these are different in PySide and PyQt
    """
    try:
        QtCoreModule = importlib.import_module('.QtCore', package=qt_module.value)
        QtGuiModule = importlib.import_module('.QtGui', package=qt_module.value)
        QtWidgetsModule = importlib.import_module('.QtWidgets', package=qt_module.value)

        if qt_module in [_QT_BINDING.PySide2, _QT_BINDING.PySide6]:
            QtCoreModule.pyqtSignal = QtCoreModule.Signal
            QtCoreModule.pyqtSlot = QtCoreModule.Slot

        return (QtCoreModule, QtGuiModule, QtWidgetsModule)
    except ImportError:
        raise ImportError(f'{qt_module} is not installed or could not be imported.')



import fcntl

import videodev2

from picamera2.platform import Platform, get_platform

from .encoder import Encoder, Quality
from .jpeg_encoder import JpegEncoder
from .libav_h264_encoder import LibavH264Encoder
from .libav_mjpeg_encoder import LibavMjpegEncoder
from .multi_encoder import MultiEncoder

_hw_encoder_available = get_platform() == Platform.VC4

if _hw_encoder_available:
    from .h264_encoder import H264Encoder
else:
    from .libav_h264_encoder import LibavH264Encoder as H264Encoder

if _hw_encoder_available:
    from .mjpeg_encoder import MJPEGEncoder
else:
    from .libav_mjpeg_encoder import LibavMjpegEncoder as MJPEGEncoder



"""This is a base class for a multi-threaded software encoder."""

import collections
import time
from fractions import Fraction
from math import sqrt

import av

import picamera2.platform as Platform
from picamera2.encoders.encoder import Encoder, Quality

from ..request import MappedArray


class LibavH264Encoder(Encoder):
    """Encoder class that uses libx264 for h.264 encoding."""

    def __init__(self, bitrate=None, repeat=True, iperiod=30, framerate=30, qp=None, profile=None):
        """Initialise"""
        super().__init__()
        self._codec = "h264"  # for now only support h264
        self.repeat = repeat
        self.bitrate = bitrate
        self.iperiod = iperiod
        self.framerate = framerate
        self.qp = qp
        self.profile = profile
        self.preset = None
        self.drop_final_frames = False
        self.threads = 0  # means "you choose"
        self._lasttimestamp = None
        self._use_hw = False
        self._request_release_delay = 1
        self._request_release_queue = None
        self._key_frames_requested = 0
        self._key_frames_generated = 0

    @property
    def use_hw(self):
        """Whether hardware encode will be used (can be set to True only for VC4 platforms)."""
        return self._use_hw

    @use_hw.setter
    def use_hw(self, value):
        """Set this property in order to get libav to use the V4L2 hardware encoder (VC4 platforms only)."""
        if value:
            if Platform.get_platform() == Platform.Platform.VC4:
                self._use_hw = True
                self._codec = "h264_v4l2m2m"
            else:
                print("Warning: use_hw has no effect on non-VC4 platforms")

    def _setup(self, quality):
        # If an explicit quality was specified, use it, otherwise try to preserve any bitrate/qp
        # the user may have set for themselves.
        if quality is not None or \
           (getattr(self, "bitrate", None) is None and getattr(self, "qp", None) is None):
            quality = Quality.MEDIUM if quality is None else quality
            # These are suggested bitrates for 1080p30 in Mbps
            BITRATE_TABLE = {Quality.VERY_LOW: 3,
                             Quality.LOW: 4,
                             Quality.MEDIUM: 7,
                             Quality.HIGH: 10,
                             Quality.VERY_HIGH: 14}
            reference_complexity = 1920 * 1080 * 30
            actual_complexity = self.width * self.height * getattr(self, "framerate", 30)
            reference_bitrate = BITRATE_TABLE[quality] * 1000000
            self.bitrate = int(reference_bitrate * sqrt(actual_complexity / reference_complexity))

    def _start(self):
        self._container = av.open("/dev/null", "w", format="null")
        self._stream = self._container.add_stream(self._codec, rate=self.framerate)

        self._stream.codec_context.thread_count = self.threads
        self._stream.codec_context.thread_type = av.codec.context.ThreadType.FRAME  # noqa

        self._stream.width = self.width
        self._stream.height = self.height
        self._stream.pix_fmt = "yuv420p"

        for out in self._output:
            out._add_stream(self._stream, self._codec, rate=self.framerate, width=self.width, height=self.height)

        preset = "ultrafast"
        if self.profile is not None:
            if not isinstance(self.profile, str):
                raise RuntimeError("Profile should be a string value")
            # Find the right profile name, ignoring case.
            profile = None
            for available_profile in self._stream.profiles:
                if self.profile.lower() == available_profile.lower():
                    profile = available_profile
                    break
            if not profile:
                raise RuntimeError("Profile " + self.profile + " not recognised")
            self._stream.profile = profile
            # The "ultrafast" preset always produces baseline, so:
            if "baseline" not in profile.lower():
                preset = "superfast"

        if self.bitrate is not None:
            self._stream.codec_context.bit_rate = self.bitrate
        # Treat gop_size 0 as meaning no I-frames at all (apart from the very first).
        self._stream.codec_context.gop_size = self.iperiod or (1 << 31) - 1

        # For those who know what they're doing, let them override the "preset".
        if self.preset:
            preset = self.preset
        self._stream.codec_context.options["preset"] = preset

        self._stream.codec_context.options["deblock"] = "1"
        # Absence of the "global header" flags means that SPS/PPS headers get repeated.
        if not self.repeat:
            self._stream.codec_context.flags |= av.codec.context.Flags.GLOBAL_HEADER  # noqa
        if self.qp is not None:
            self._stream.codec_context.qmin = self.qp
            self._stream.codec_context.qmax = self.qp

        self._stream.codec_context.time_base = Fraction(1, 1000000)
        self._stream.codec_context.options["tune"] = "zerolatency"

        FORMAT_TABLE = {"YUV420": "yuv420p",
                        "BGR888": "rgb24",
                        "RGB888": "bgr24",
                        "XBGR8888": "rgba",
                        "XRGB8888": "bgra"}
        self._av_input_format = FORMAT_TABLE[self._format]

        self._request_release_queue = collections.deque()

    def _stop(self):
        if not self.drop_final_frames:
            # Annoyingly, libav still has lots of encoded frames internally which we must flush
            # out. If the output(s) doesn't understand timestamps, we may need to "pace" these
            # frames with correct time intervals. Unpleasant.
            for packet in self._stream.encode():
                if any(out.needs_pacing for out in self._output) and self._lasttimestamp is not None:
                    time_system, time_packet = self._lasttimestamp
                    delay_us = packet.pts - time_packet - (time.monotonic_ns() - time_system) / 1000
                    if delay_us > 0:
                        time.sleep(delay_us / 1000000)
                self._lasttimestamp = (time.monotonic_ns(), packet.pts)
                self.outputframe(bytes(packet), packet.is_keyframe, timestamp=packet.pts, packet=packet)
        while self._request_release_queue:
            self._request_release_queue.popleft().release()
        self._container.close()

    def _encode(self, stream, request):
        request.acquire()
        self._request_release_queue.append(request)
        timestamp_us = self._timestamp(request)
        with MappedArray(request, stream) as m:
            frame = av.VideoFrame.from_numpy_buffer(m.array, format=self._av_input_format, width=self.width)
            frame.pts = timestamp_us
            if self._key_frames_requested > self._key_frames_generated:
                self._key_frames_generated += 1
                frame.pict_type = "I"
            for packet in self._stream.encode(frame):
                self._lasttimestamp = (time.monotonic_ns(), packet.pts)
                self.outputframe(bytes(packet), packet.is_keyframe, timestamp=packet.pts, packet=packet)
        while len(self._request_release_queue) > self._request_release_delay:
            self._request_release_queue.popleft().release()

    def force_key_frame(self):
        """Force a key frame to be encoded in the video stream as soon as possible."""
        self._key_frames_requested += 1



import logging

from picamera2.allocators import DmaAllocator

_log = logging.getLogger("picamera2")


class PersistentAllocator(DmaAllocator):
    """Persistent DmaHeap Allocator"""

    def __init__(self):
        super().__init__()
        self.buffer_key = None
        self.buffer_dict = {}

    def allocate(self, libcamera_config, use_case):
        if use_case is None:
            _log.error("Must set use_case before using persistent allocator")
        self.buffer_key = use_case

        self.open_fds = []
        self.libcamera_fds = []
        self.frame_buffers = {}
        self.mapped_buffers = {}
        self.mapped_buffers_used = {}

        buffers = self.buffer_dict.get(self.buffer_key)
        if buffers is None:
            super().allocate(libcamera_config, use_case)
            self.buffer_dict[self.buffer_key] = (
                self.open_fds,
                self.libcamera_fds,
                self.frame_buffers,
                self.mapped_buffers,
                self.mapped_buffers_used,
            )
        else:
            (self.open_fds, self.libcamera_fds, self.frame_buffers,
             self.mapped_buffers, self.mapped_buffers_used) = buffers

    def cleanup(self):
        pass

    def deallocate(self, buffer_key=None):
        """Deallocate a set of buffers if no longer in use"""
        if buffer_key is None:
            buffer_key = self.buffer_key

        tmp = super().__new__(DmaAllocator)
        tmp.dmaHeap = None

        (tmp.open_fds, tmp.libcamera_fds, tmp.frame_buffers,
         tmp.mapped_buffers, tmp.mapped_buffers_used) = self.buffer_dict[buffer_key]
        tmp.close()
        del self.buffer_dict[buffer_key]

    def close(self):
        for k in list(self.buffer_dict.keys()):
            self.deallocate(k)
        assert self.buffer_dict == {}



"""
Nanodet postprocessing

This code is based on:
https://github.com/RangiLyu/nanodet
"""

from typing import Tuple

import numpy as np

from picamera2.devices.imx500.postprocess import combined_nms, softmax


def postprocess_nanodet_detection(outputs,
                                  conf: float = 0.0,
                                  iou_thres: float = 0.65,
                                  max_out_dets: int = 300) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    reg_max = 7
    num_categories = 80
    classes = outputs[..., :num_categories]
    boxes = outputs[..., num_categories:]
    classes = 1 / (1 + np.exp(-classes))  # sigmoid

    # Extract feature map sizes
    strides = [8, 16, 32, 64]
    featmap_sizes = [(np.ceil(416 / stride), np.ceil(416 / stride)) for stride in strides]

    # Generate priors
    anchors = generate_anchors_NANODET(featmap_sizes, strides)

    # Decode bboxes
    batch = boxes.shape[0]
    x = np.reshape(boxes, newshape=(batch, -1, 4, reg_max + 1))
    x = softmax(x)
    x = np.matmul(x, np.arange(0, reg_max + 1, 1, dtype=np.float32))
    x = np.reshape(x, newshape=(batch, -1, 4))
    distances = x * anchors[..., 2, None]

    # Output Box format: [x_c, y_c, w, h]
    w = distances[..., 0:1] + distances[..., 2:3]
    h = distances[..., 1:2] + distances[..., 3:4]
    x_c = anchors[..., 0:1] - distances[..., 0:1] + w / 2
    y_c = anchors[..., 1:2] - distances[..., 1:2] + h / 2
    boxes = np.concatenate([x_c, y_c, w, h], axis=2)

    return combined_nms(boxes, classes, iou_thres, conf, max_out_dets)


def generate_anchors_NANODET(featmap_sizes, strides):
    anchors_list = []
    for i, stride in enumerate(strides):
        h, w = featmap_sizes[i]
        x_range = np.arange(w) * stride
        y_range = np.arange(h) * stride
        y, x = np.meshgrid(y_range, x_range)
        y = y.flatten()
        x = x.flatten()
        strides = np.ones_like(x) * stride
        anchors = np.stack([y, x, strides, strides], axis=-1)
        anchors = np.expand_dims(anchors, axis=0)
        anchors_list.append(anchors)
    return np.concatenate(anchors_list, axis=1)



from .circularoutput import CircularOutput
from .circularoutput2 import CircularOutput2
from .ffmpegoutput import FfmpegOutput
from .fileoutput import FileOutput
from .output import Output
from .pyavoutput import PyavOutput
from .splittableoutput import SplittableOutput



# Installing Qt and OpenGL on the 64-bit Lite OS, and trying to run
# a remote preview window, causes an error here (the 32-bit Lite OS is OK).
# It may be something to do with more recent versions of python3-opengl?
# Anyway, if we carry on regardless at least the non-OpenGL preview works,
# which is in any case what is required for remote preview windows.
from logging import getLogger

from .q_picamera2 import _get_qpicamera2
from .qt_compatibility import _QT_BINDING

_log = getLogger(__name__)

try:
    from .q_gl_picamera2 import _get_qglpicamera2
except Exception:
    _log.warning("OpenGL will not be available")


# Lazy load QPicamera2 widget classes as will likely only use one or two within a given application
def __getattr__(name: str):
    # Standard Qt widgets
    if name == 'QPicamera2':
        return _get_qpicamera2(_QT_BINDING.PyQt5)
    elif name == 'Q6Picamera2':
        return _get_qpicamera2(_QT_BINDING.PyQt6)
    elif name == 'QSide2Picamera2':
        return _get_qpicamera2(_QT_BINDING.PySide2)
    elif name == 'QSide6Picamera2':
        return _get_qpicamera2(_QT_BINDING.PySide6)
    # OpenGL accelerated Qt widgets
    elif name == 'QGlPicamera2':
        return _get_qglpicamera2(_QT_BINDING.PyQt5)
    elif name == 'QGl6Picamera2':
        return _get_qglpicamera2(_QT_BINDING.PyQt6)
    elif name == 'QGlSide2Picamera2':
        return _get_qglpicamera2(_QT_BINDING.PySide2)
    elif name == 'QGlSide6Picamera2':
        return _get_qglpicamera2(_QT_BINDING.PySide6)
    raise AttributeError(f"qt has no attribute '{name}'")
